

var showReductions = true;

//needed for free variable lists - removes ele from array
function remove (ele, arr) {
  return arr.filter(function (a) { return a !== ele; });
}

//creates a fresh variable name
var gensymcount = 0;
function gensym () {
  gensymcount = gensymcount + 1;
  return "#x" + gensymcount;
}

//lambda prototype for function abstraction
function Lambda (binder, body) {
  this.binder = binder;
  this.body = body;
}

//toString - prints λ binder → body
//reduce - reduces the body, and creates a new abstraction with the reduced body
//substitute - if old is binder, everything is done
//           - if new is free in body - create fresh name, substitute binder for fresh, and afterwards new for old
//           - else substitute in body
//free - FV(λ x → e) = FV(e) \ x
//binder keeps the name of the binder
//body keeps the body term
Lambda.prototype.toString =
  function () {
    return "λ " + this.binder + " → " + this.body.toString();
  };
Lambda.prototype.reduce =
  function () { return new Lambda(this.binder, this.body.reduce()); };
Lambda.prototype.substitute =
  function (o, n) {
    if (this.binder === o) //Verschattung
      return this;
    else if (n.free().indexOf(this.binder) !== -1) { //free ist später implementiert
      //Bezeichner mit gleichem Namen, also zuerst einen frischen erzeugen
      var ns = gensym(); //gensym ist später implementiert
      //und die gebundene Variable zuerst ersetzen
      return new Lambda(ns, this.body.substitute(this.binder, new Variable(ns)).substitute(o, n));
    } else
      //Normalfall: Delegation der Ersetzung an den Funktionsrumpf
      return new Lambda(this.binder, this.body.substitute(o, n));
  };
Lambda.prototype.free =
  function () { return remove(this.binder, this.body.free()); };

//prototype for applications

function Application (left, right) {
  this.left = left;
  this.right = right;
}

//toString - prints left side and right side, separated by a whitespace
//reduce - reduces right hand side (argument), reduces left hand side
//         if left hand is a lambda, substitute binder with reduced right hand side
//         afterwards reduces result
//substitute - applies substitution on the left and on the right
//free - FV(M N) = FV(M) u FV(N)
//right and left contain the terms of right and left hand side
Application.prototype.toString =
  function () {
    var l = this.left.toString();
    var r = this.right.toString();
    if (this.left instanceof Lambda)
      l = "(" + l + ")";
    if (this.right instanceof Lambda || this.right instanceof Application)
      r = "(" + r + ")";
    return l + " " + r;
  };
Application.prototype.reduce =
  function () {
    var nr = this.right.reduce();
    var nl = this.left.reduce();
    if (nl instanceof Lambda) {
      var r = nl.body.substitute(nl.binder, nr);
      if (showReductions)
        console.log(new Application(nl, nr).toString() + " ⟿  " + r.toString());
      var s = r.reduce();
      return s;
    } else
      return new Application(nl, nr);
  };
Application.prototype.substitute =
  function (o, n) {
    return new Application(this.left.substitute(o, n), this.right.substitute(o, n));
  };
Application.prototype.free =
  function () { return this.left.free().concat(this.right.free()); };


function Variable (name) {
  this.name = name;
}

//toString - just its name
//reduce - no operation
//substitute - if it is us to be replaced, we replace us
//free - ourselves (the name), as an array
Variable.prototype.toString =
  function () { return this.name; };
Variable.prototype.reduce = function () { return this; };
Variable.prototype.substitute =
  function (o, n) {
    if (this.name === o)
      return n;
    else
      return this;
  };
Variable.prototype.free = function () { return [this.name]; };



//should live somewhere else....


//zero - the number zero, in church encoding λ f → (λ x → x)
var z = new Lambda("s", new Lambda("z", new Variable("z")))

// conversion from JS numbers to church numbers
//church :: Integer -> Church Integer
//church 0 = \f -> \x -> x
//church n = \f -> \x -> f (church (n-1) f x)

function toChurch (n) {
  if (n === 0)
    return z
  else {
    var rec = toChurch(n - 1)
    return new Lambda("s", new Lambda("z", new Application(new Variable("s"), new Application(new Application(rec, new Variable("s")), new Variable("z"))))).reduce()
  }
}

//looks neat on paper, but reality is different
//namely, no arbitrary JS expressions in our lambdas!
//unchurch :: Church Integer -> Integer
//unchurch n = n (\x -> x + 1) 0
//luckily we can count the amount of Applications
function toJS (church) {
  if (church instanceof Lambda) { // ist church eine Funktionsabstraktion?
    var name = church.binder;
    if (church.body instanceof Lambda) { // ist church.body eine Funktionsabstraktion?
      return countNestedApplications(church.body.body, name);
    }
  }
}

function countNestedApplications (expression, binder) {
  if ((expression instanceof Application) && expression.left.name === binder)
    return 1 + countNestedApplications(expression.right, binder);
  else
    return 0;
}



